/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_DATA_STORAGE_HPP
#define FGGL_DATA_STORAGE_HPP

#include <iostream>
#include <string>
#include <filesystem>
#include <utility>
#include <vector>

#include "fggl/debug/logging.hpp"
#include "fggl/modules/module.hpp"
#include "fggl/platform/paths.hpp"

namespace fggl::data {

	template<typename T>
	bool fggl_serialize(std::filesystem::path &data, const T *out);

	template<typename T>
	bool fggl_deserialize(std::filesystem::path &data, T *out);

	enum StorageType { Data, Config, Cache };

	constexpr const modules::ModuleService SERVICE_STORAGE = modules::make_service("fggl::data::Storage");

	class Storage {
		public:
			constexpr static modules::ModuleService service = SERVICE_STORAGE;
			Storage( fggl::platform::EnginePaths paths ) : m_paths(std::move( paths )) {}

			template<typename T>
			bool load(StorageType pool, const std::string &name, T *out) {
				auto path = resolvePath(pool, name);
				if (!std::filesystem::exists(path)) {
					debug::log(debug::Level::warning, "Attempted to load '{}', but it did not exist", path.string());
					return false;
				}
				return fggl_deserialize<T>(path, out);
			}

			std::vector<std::filesystem::path> findResources(std::filesystem::path root, const std::string ext) {
				if ( !std::filesystem::exists(root) ) {
					return {};
				}

				std::vector<std::filesystem::path> paths;
				for ( const auto& entry : std::filesystem::recursive_directory_iterator(root) ) {
					if ( entry.is_regular_file() && entry.path().extension() == ext ) {
						paths.push_back( entry );
					}
				}
				return paths;
			}

			template<typename T>
			void save(StorageType pool, const std::string &name, const T *out) {
				auto path = resolvePath(pool, name, true);
				fggl_serialize<T>(path, out);
			}

			inline std::filesystem::path resolvePath(StorageType pool, const std::string &name, bool createParents = false) {
				std::filesystem::path path;
				switch (pool) {
					case Data:
						path = fggl::platform::locate_data(m_paths, name);
						break;
					case Config:
						path = fggl::platform::locate_config(m_paths, name);
						break;
					case Cache:
						path = fggl::platform::locate_cache(m_paths, name);
						break;
				}

				if ( createParents ){
					if ( !std::filesystem::exists(path.parent_path()) ) {
						std::filesystem::create_directories(path.parent_path());
					}
				}

				return path;
			}

		private:
			fggl::platform::EnginePaths m_paths;
	};

}

#endif
