/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_MATH_SHAPES_HPP
#define FGGL_MATH_SHAPES_HPP

#include "fggl/math/types.hpp"
#include <vector>
#include <array>

// based on https://gamemath.com/book/geomprims.html
namespace fggl::math::phs3d {

	struct Line {
		math::vec2 normal;
		float d;
	};

	struct LineSegment {
		math::vec3 start;
		math::vec3 end;
	};

	struct Ray {
		math::vec3 start;
		math::vec3 length;

		inline math::vec3 end() {
			return start + length;
		}

		static Ray fromPos(math::vec3 start, math::vec3 end);
	};

	// nice bounding shapes

	struct Sphere {
		math::vec3 center;
		float radius;

		float volume() const;
		float area() const;
	};

	struct AABB {
		math::vec3 min;
		math::vec3 max;

		void add(const math::vec3& p);
		void emtpy();
		void set(const AABB& other, const math::mat4& m);

		inline math::vec3 center() const {
			return (min + max) / 2.0F;
		}

		inline math::vec3 size() const {
			return max - min;
		}

		inline math::vec3 radius() const {
			return max - center();
		}

		static AABB fromPoints(const std::vector<math::vec3>& points);
	};

	struct Plane {
		glm::vec3 normal;
		float d; // distance to origin

		inline bool contains(const math::vec3& point) {
			return glm::dot(point, normal) == d;
		}

		inline float distance(const math::vec3& q) {
			return glm::dot(q, normal) - d;
		}

		static Plane fromPoints(const math::vec3, const math::vec3, const math::vec3);
		static Plane bestFit(const std::vector<math::vec3>& points);
	};

	struct Barycentric {

		std::array<float, 3> b;

		bool inTriangle() {
			return 0 <= b[0] && b[0] <= 1 &&
				   0 <= b[1] && b[1] <= 1 &&
				   0 <= b[2] && b[2] <= 1;
		}

		inline bool isLegal() const {
			return (b[0] + b[1] + b[2]) == 1.0f;
		}
	};

	constexpr float THIRD = 1.0F/3.0F;
	const Barycentric cGrav = { {THIRD, THIRD, THIRD}};

	struct Triangle {
		std::array<math::vec3, 3> v;

		inline float l1() const {
			return length(2, 1);
		}

		inline float l2() const {
			return length(0, 2);
		}

		inline float l3() const {
			return length(1, 0);
		}

		inline math::vec3 edge(int v1, int v2) const {
			return v[v1] - v[v2];
		}

		inline float length(int a, int b) const {
			return glm::length( edge(a, b));
		}

		inline float perimeter() const {
			return length(3, 2) + length(1,3) + length(2, 1);
		}

		inline float area() const {
			return glm::length(glm::cross(edge(3, 2), edge(1, 3))) / 2;
		}

		inline math::vec3 cGrav() const {
			return (v[0] + v[1] + v[2]) / 3.0f;
		}

		inline math::vec3 inCenter() const {
			auto p = perimeter();
			Barycentric bs{
				l1() / p,
				l2() / p,
				l3() / p
			};
			return BarycentricToCart(bs);
		}

		inline float centerRadius() const {
			return area() / perimeter();
		}

		// circle that passes through all three points = cirumcenter + circumraidus
		inline math::vec3 circumcenter() {
			// d1 = - e2 DOT e3
			// d2 = - e3 DOT e1
			// d3 = - e1 DOT e2
			// C1 = d2 * d3
			// C2 = d3 * d1
			// C3 = d1 * d2
			// c = c1 + c2 + c3
			// c_2 = 2 * c
			// bary{
			//   (c2 + c3 ) / c_2
			//   (c3 + c1 ) / c_2
			//   (c1 + c2 ) / c_2
			// return BarycentricToCart(bary)
			return glm::zero<vec3>();
		}

		inline float circumradius() {
			// d1 = - e2 DOT e3
			// d2 = - e3 DOT e1
			// d3 = - e1 DOT e2
			// C1 = d2 * d3
			// C2 = d3 * d1
			// C3 = d1 * d2
			// c = c1 + c2 + c3
			// c_2 = 2 * c
			// return sqrt( (d1+d2) * (d2+d3) * ( d3 + d1 ) / c ) / 2
			return 0.0F;
		}

		bool CartToBarycentric(const math::vec3& cart, Barycentric& outVal);
		bool CartToBarycentric2(const math::vec3& cart, Barycentric& outVal);

		math::vec3 BarycentricToCart(const Barycentric& p) const {
			return v[0] * p.b[0] +
				   v[1] * p.b[1] +
				   v[2] * p.b[2];
		}
	};

} // namespace fggl::math::phys3d


#endif //FGGL_MATH_SHAPES_HPP
