/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_MATH_TRIANGULATION_HPP
#define FGGL_MATH_TRIANGULATION_HPP

#define _USE_MATH_DEFINES
#include <cmath>

#include <fggl/math/types.hpp>
#include <fggl/data/model.hpp>

namespace fggl::math {

	using Polygon = std::vector<math::vec2>;
	using PolygonVertex = std::vector<data::Vertex2D>;

	constexpr int POSITIVE = 1;
	constexpr int NEGATIVE = -1;
	constexpr int UNSET = 0;

	/**
	 * Put an angle in the range [-PI, PI].
	 */
	inline float clampAngle(float radianAngle) {
		if (radianAngle <= math::PI) {
			return radianAngle + math::HALF_PI;
		} else if (radianAngle > math::PI) {
			return radianAngle - math::HALF_PI;
		} else {
			return radianAngle;
		}
	}

	static void checkSign(float value, int &sign, int &firstSign, int &flips) {
		if (value > 0) {
			if (sign == UNSET) {
				firstSign = POSITIVE;
			} else if (sign < 0) {
				flips++;
			}
		} else if (value < 0) {
			if (sign == UNSET) {
				firstSign = NEGATIVE;
			} else if (sign > 0) {
				flips++;
			}
			sign = NEGATIVE;
		}
	}

	/**
	 * Check if a polygon is convex.
	 *
	 * see https://math.stackexchange.com/a/1745427
	 */
	static bool isConvex(const Polygon &polygon) {
		if (polygon.size() < 3) {
			return false;
		}

		const auto n = polygon.size();

		auto wSign = UNSET;

		auto xSign = UNSET;
		auto xFirstSign = UNSET;
		auto xFlips = 0;

		auto ySign = UNSET;
		auto yFirstSign = UNSET;
		auto yFlips = 0;

		auto curr = polygon[n - 1];
		auto next = polygon[n];

		for (auto &v : polygon) {
			auto prev = curr;
			curr = next;
			next = v;

			auto before = curr - prev;
			auto after = next - curr;

			checkSign(after.x, xSign, xFirstSign, xFlips);
			if (xFlips > 2) {
				return false;
			}

			checkSign(after.y, ySign, yFirstSign, yFlips);
			if (yFlips > 2) {
				return false;
			}

			auto w = before.x * after.y - after.x * before.y;
			if (wSign == UNSET && w != 0) {
				wSign = w;
			} else if (wSign > 0 && w < 0) {
				return false;
			} else if (wSign < 0 && w > 0) {
				return false;
			}
		}

		if (xSign != UNSET && (xFirstSign != UNSET) && (xSign != xFirstSign)) {
			xFlips += 1;
		}
		if (ySign != UNSET && (yFirstSign != UNSET) && (ySign != yFirstSign)) {
			yFlips += 1;
		}
		if (xFlips != 2 || yFlips != 2) {
			return false;
		}

		return true;
	}

	static data::Vertex2D pointToVertex(const math::vec2 &point) {
		return data::Vertex2D{ point, {1.0f, 1.0f, 1.0f} };
	}

	/**
	 * Fast Triangulation for convex polygons.
	 */
	void fan_triangulation(const PolygonVertex& polygon, data::Mesh2D &mesh);

} // namespace fggl::util

#endif

