/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 01/08/22.
//

#include "fggl/phys/bullet/service.hpp"
#include "fggl/phys/bullet/types.hpp"
#include "fggl/phys/bullet/motion.hpp"

namespace fggl::phys::bullet {

	constexpr float UNIT_EXTENT = 0.5F;

	btCollisionShape* create_bullet_shape(const YAML::Node& node) {
		auto type = node["type"].as<std::string>("MISSING");
		if ( type == "sphere" ) {
			auto radius = node["radius"].as<float>(UNIT_EXTENT);
			return new btSphereShape(radius);
		} else if ( type == "box" ) {
			auto extents = node["extents"].as<math::vec3>(math::vec3(UNIT_EXTENT));
			return new btBoxShape(btVector3(extents.x, extents.y, extents.z));
		} else {
			debug::warning("unknown shape type: {}, assuming unit sphere", type);
			return new btSphereShape(0.5F);
		}
	}

	void setup_body(BodyType type, btRigidBody& body) {
		if ( type != BodyType::STATIC ) {
			body.setRestitution(0.5F);
			body.setRollingFriction(0.0F);

			// kinematic bodies cannot sleep
			if (type == BodyType::KINEMATIC) {
				auto flags = (body.getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT);
				body.setCollisionFlags(flags);
				body.setActivationState(DISABLE_DEACTIVATION);
			}
		}
	}

	void add_bt_body(const entity::ComponentSpec& spec, entity::EntityManager& manager, const entity::EntityID& id) {
		auto& body = manager.add<BulletBody>(id);
		body.motion = new FgglMotionState(&manager, id);

		auto type = spec.get<BodyType>("type", BodyType::DYNAMIC);
		auto mass = 0.0F;

		// non-static objects can have mass
		if ( type != BodyType::STATIC ) {
			mass = spec.get<float>("mass", 1.0F);
		}

		// calculate inertia
		btCollisionShape* shape = create_bullet_shape(spec.config["shape"]);
		btVector3 inertia(0, 0, 0);
		if ( type != BodyType::STATIC ) {
			shape->calculateLocalInertia(mass, inertia);
		}

		// construction information for the simulation
		btRigidBody::btRigidBodyConstructionInfo constructionInfo {
			mass,
			body.motion,
			shape,
			inertia
		};

		body.body = new btRigidBody(constructionInfo);
		setup_body(type, *body.body);

		// add the body to the simulation
		if ( type == BodyType::DYNAMIC ) {
			manager.add<phys::Dynamics>(id);
		}
	}

	PhysicsEngine* BulletProvider::create(entity::EntityManager* entityManager, entity::EntityFactory* factory) {
		auto* engine = new BulletPhysicsEngine(entityManager);

		// FIXME: this won't work across multiple scenes - reactive approach might be better...
		factory->bind(util::make_guid("phys::Body"), add_bt_body, [engine](auto& manager, const auto& entity) {
			debug::trace("adding entity {} to physics scene...", (uint64_t)entity);
			auto& physComp = manager.template get<BulletBody>(entity);
			engine->addBody(entity, physComp);
		});

		return engine;
	}

} // namespace fggl::phys::bullet